{-

Open Source Fixedpoint Model-Checker version 2010

(C) Copyright IBM Corp. 2010

All Rights Reserved.

Modifications by Paolo Modesti 2010

-}

module Main
where
import FPASLanLexer
import FPASLanParser
import System
import Data.Char
import Data.List
import Data.Maybe
import qualified Data.Map as Map

debug = False
spassout = True
main = 
  do args <- System.getArgs
     file <- case args of
             [filename] -> readFile filename
	     [filename,prov] -> readFile filename
	     _ -> error "Specify Input File"
     spassout <- case args of 
     	          [filename] -> return True
		  [filename,prov] -> if prov=="-proverif" then return False
		  		     else error $ "Unrecognized option "++prov
     (paslan@(name,types,sets,pubfun,privfun,facts,definitions,rules)) <- return . mkPrepros . parser . alexScanTokens $ file
     -- <paolo>--
     if debug then error . dbgDefs . parser . alexScanTokens $ file else putStr ""
     -- </paolo>--
     symbtab <- return $ mkSymbtab paslan
     -- putStr $ Map.showTree symbtab
     putStr $ if spassout then "% Checking rules: "++(show (checkrules symbtab rules))++"\n" else "(* Checking rules: "++(show (checkrules symbtab rules))++"*)\n"
     (rules',numsets) <- return $ absRules symbtab (quantelims symbtab rules)
     putStr ((if spassout then pASLan else prASLan) symbtab numsets name rules')

type Symbtab = Map.Map String SymbEntry

data SymbEntry = EnuCo 
     	       | VarEnu [String]
	       | VarVal
	       | ConVal
	       | VarUntyped
	       | FunSym Bool Int
	       | FactSym Int
	       | SetSym [String]
	       deriving (Eq,Show)

isEnu (VarEnu _) = True
isEnu _ = False
isVal VarVal = True
isVal ConVal = True
isVal _ = False
isSet (SetSym _) = True
isSet _ = False
isFun :: SymbEntry -> Bool
isFun (FunSym _ _) = True
isFun _ = False
isFact (FactSym _) = True
isFact _ = False
isEnuCo :: SymbEntry -> Bool
isEnuCo EnuCo = True
isEnuCo _ = False
isPub (FunSym True _) = True
isPub _ = False
arity :: SymbEntry -> Int
arity (FunSym _ n) = n
arity (FactSym n)  = n 
arity EnuCo = 0
arity (VarEnu _) = 0
arity (VarVal) = 0
arity ConVal =0
arity VarUntyped =0
arity (SetSym l) = length l

mkSymbtab :: PASLan -> Symbtab
mkSymbtab (_,types,sets,pubfun,privfun,facts,definitions,rules) =

  ((mkSTfac facts) .
  (mkSTfun pubfun privfun) .
  (mkSTsets sets))
  (mkSTtypes types)

builtinSym = 
 [("0",FunSym True 0),
  ("1",FunSym True 0),
  ("timplies",FactSym 2),
  --- ("val",FunSym False 0),
  ("epsilon",FunSym True 0)
  --("occurs",FactSym 1)
 ]

mkSTtypes types = 
  foldr mkSTtype (Map.fromList builtinSym) types

vinsert :: String -> SymbEntry -> Symbtab -> Symbtab
vinsert k v m = if Map.member k m then error $ "Multiple Decs of "++k
	        else Map.insert k v m

mapfold :: Symbtab -> [(String,SymbEntry)] -> Symbtab
mapfold = foldr (\ (k,v) tab -> vinsert k v tab)

mkSTtype (vars,Enum ids) table =
  let checkEC key old new = 
      	if isVar key then error $ key ++ " is a variable (used in enum)" else
        if new==EnuCo then EnuCo 
      	else error $ key++" used in enumeration and in some other way"
      tab'=foldr (\ id tab -> Map.insertWithKey checkEC id EnuCo tab) 
      		 table ids
  in mapfold tab' (map (\ v -> if isVar v then (v,(VarEnu ids)) 
     	     	       	       else error $ v ++ " is not a variable name") 
		       vars)
mkSTtype (vars,Value) table = 
  mapfold table 
  (map (\ id -> if isVar id then (id,VarVal) else (id,ConVal)) vars)
mkSTtype (vars,Untyped) table = 
  mapfold table
  (map (\ id -> (id,VarUntyped)) vars)

mkSTsets sets table = foldr mkSTset table sets

mkSTset :: PTerm -> Symbtab -> Symbtab
mkSTset (PComp setname args) table =
  let checkname n = let r = Map.lookup n table
      		    in if isJust r then 
		       	  if isEnu $ fromJust r then n
		       	  else error $ n ++ " is not an enumeration type" 
                       else error $ n ++ " is undefined "
      getname (PAtom m) = checkname m
      getname m = error $ (show m) ++ " as argument in set description"
  in if isVar setname then error $ setname ++ " is not a setname"
     else vinsert setname (SetSym (map getname args)) table
mkSTset term _ = error $ (show term)++" is not a set description"

mkSTfun pub priv table = 
  mapfold table
  ((map (\ (f,n) -> if isVar f then error $ f++" is not a function symbol"
  	   	    else (f,FunSym True n)) pub)++
   (map (\ (f,n) -> if isVar f then error $ f++" is not a function symbol"
  	   	    else (f,FunSym False n)) priv))

mkSTfac facts table = 
  mapfold table
  (map (\ (f,n) -> if isVar f then error $ f++" is not a fact symbol"
  	   	    else (f,FactSym n)) facts)
isVar = isUpper . head

------------------------------

chexpl :: String -> Bool -> Bool
chexpl s b = if b then b else error s

checkrules :: Symbtab -> [PRule] -> Bool
checkrules tab rules =
  all (checkrule tab) rules

checkrule tab (lambda,lf,ps,ns,f,rf,rs) =
  let defs0 = nub $ (fvfs lf)++(fvss ps)
      defs  = nub $ lambda++f++(filter (not . isEnu . ((!) tab)) defs0)
      uses  = nub $ (fvfs rf)++(fvss rs)++(fvss ns)++(filter (isEnu . ((!) tab)) defs0) in
  (all (checkDefEnu tab) lambda) &&
  (all (checkFact tab) lf) &&
  (all (checkFact tab) rf) &&
  (all (checkCond tab True ) ps) &&
  (all (checkCond tab False) ns) &&
  (all (checkCond tab True ) rs) &&
  (chexpl ("using undefined vars: "++(show (uses\\defs))) $ null (uses \\ defs)) &&
  (chexpl ("creating defined var") $ null (defs0 `intersect` (nub f)))

fvfs = nub . (concatMap fvf)
fvf (fsym,ts) = fvts ts

fvts = nub . (concatMap fvt)
fvt (PAtom p) = if isVar p then [p] else []
fvt (PComp f ts) = fvts ts
--  defintions
fvt (PParId p) = error $ "error in definitions translation - Illegal PParId "++p
fvt (PPar f ts) = error $ "error in definitions translation - Illegal PPar "++f
-- definitions
fvss = nub . (concatMap fvs)
fvs (uni,t,s) = (nub $ (if isVar t then [t] else [])++(fvt s)) \\ uni

(!) = \ m k -> fromMaybe (error $ k++" undefined") $ Map.lookup k m

checkFact :: Symbtab  -> PFact -> Bool
checkFact tab (f,ts) = 
  case tab ! f of
  (FactSym n) -> (chexpl ("arity of "++f) $ (n==(length ts))) && (all (checkTerm tab) ts)
  _ -> error $ f++" is not a fact symbol."

checkTerm tab (PAtom a) =
  case tab ! a of
  (FunSym _ _) -> error $ "Illegal atom "++a
  (FactSym _)  -> error $ "Illegal atom "++a
  (SetSym _) -> error $ "Illegal atom "++a
  _ -> True 
  
checkTerm tab (PComp f ts) =
  let (FunSym _ n) = tab ! f
  in (chexpl ("arity of "++f) $ (n==(length ts))) && (all (checkTerm tab) ts)

-- definitions --
checkTerm tab (PParId a) = error $ "error in definitions translation - Illegal PParId "++a
checkTerm tab (PPar f ts) = error $ "error in definitions translation - Illegal PPar "++f
--
  
checkCond tab positive (uni,t,s) =
  (if positive then chexpl "positive fact with quantifier" $ null uni 
   else all (checkDefEnu tab) uni) &&
  (checkDefVal tab t) &&
  (checkSet tab s)

checkDefEnu tab v = chexpl (v++" is not an enumeration constant") $ isEnu (tab ! v)
checkDefVal tab t = chexpl (t++" is not a value") $ isVal (tab ! t)
checkDefSet tab s = chexpl (s++" is not a set constructor") $ isSet (tab ! s)

checkSet :: Symbtab -> PTerm -> Bool 
checkSet tab (PAtom a) = error $ a++" is not a set."
checkSet tab (PComp f ts) = 
  let (SetSym args) = tab ! f
      ts' = map getIdent ts 
      getIdent (PAtom a) = a
      getIdent a = error $ (show a)++" is not a valid argument for a set term" 
  in (chexpl ("wrong length of arguments "++(show ts)) $ (length args) == (length ts)) &&
     (all (checkArg tab) (zip ts' args))

checkSet tab (PParId a) = error $ a++" is not a set."
checkSet tab (PPar f ts) = error $ f++" is not a valid argument"

checkArg :: Symbtab -> (PIdent,PIdent) -> Bool
checkArg tab (actual,formal) =
  let (VarEnu fextension) = tab ! formal
  in case tab ! actual of
     EnuCo -> chexpl (actual++" not in "++formal) $ actual `elem` fextension
     VarEnu m -> chexpl ("sets with too large extension "++actual) $ null (m\\fextension)
     _ -> error $ "illegal parameter in set term: "++actual

------- quantifier elimination

quantelims :: Symbtab -> [PRule] -> [PRule]

quantelims tab = concatMap $ quantelim tab

quantelim :: Symbtab -> PRule -> [PRule]

quantelim tab (rule@(lamvar,lf,ps,ns,f,rf,rs)) =
  let nsqf = quantelimns tab ns in
  foldr (\ var rules ->
  	   concat [subrs sigma rules | sigma <- getvals tab var ])
         [([],lf,ps,nsqf,f,rf,rs)] lamvar

getvals :: Symbtab -> PIdent -> [Substitution]
getvals tab var = 
   let VarEnu vals = tab ! var 
   in [Map.singleton var (PAtom v) |v<-vals]

quantelimns :: Symbtab -> [PCond] -> [PCond]
quantelimns tab = concatMap $ quantelimn tab

quantelimn :: Symbtab -> PCond -> [PCond]
quantelimn tab (qs,t,s) = 
  nub $ 
  foldr (\ var conds ->
           concat [subcs sigma conds | sigma <- getvals tab var])
	[([],t,s)] qs

type Substitution = Map.Map PIdent PTerm

subrs :: Substitution -> [PRule] -> [PRule]
subr  :: Substitution ->  PRule  ->  PRule
subfs :: Substitution -> [PFact] -> [PFact]
subf  :: Substitution ->  PFact  ->  PFact
subcs :: Substitution -> [PCond] -> [PCond]
subc  :: Substitution ->  PCond  ->  PCond
subts :: Substitution -> [PTerm] -> [PTerm]
subt  :: Substitution ->  PTerm  ->  PTerm

subrs sub = map $ subr sub
subr sub ([],lf,ps,ns,f,rf,rs) = 
         ([],subfs sub lf,
	     subcs sub ps, 
	     subcs sub ns,
	     f,
	     subfs sub rf, 
	     subcs sub rs)
subfs sub = map (subf sub)
subf  sub (f,ts) = (f,subts sub ts)
subcs sub = map (subc sub)
subc  sub ([],t,set) = ([],t,subt sub set)
subts sub = map (subt sub)
subt  sub (PAtom a) = Map.findWithDefault (PAtom a) a sub
subt  sub (PComp f ts) = PComp f (subts sub ts)

subt  sub (PParId a) = error ("error in definitions translation - PParId " ++ a)
subt  sub (PPar f ts) = error ("error in definitions translation - PPar " ++ f)
------- abstract

type Alpha = (PTerm -> Int)

mkAbs :: Symbtab -> [PTerm] -> (Alpha,Int)
mkAbs tab terms = 
  let (allsets::[PTerm])
              = nub $ 
      	      	foldr (\ var ts ->
		      	 concat [subts sigma ts| sigma<- getvals tab var])
                terms (fvts terms) 
  -- ERROR in SPASS OUTPUT in some cases (val,num) where num = length allsets
  -- (val,0) returns 0
  in ((Map.!) (Map.fromList (zip allsets [1..])),length allsets)

absRules :: Symbtab -> [PRule] -> ([PRule],Int)
absRules tab rules =
  let ali = mkAbs tab (getSetsrs rules)
  in  (map (\rule -> foldr (absRule tab ali) rule (absr tab rule)) rules,snd ali)

absRule :: Symbtab -> (Alpha,Int) -> PIdent -> PRule -> PRule
absRule tab (alpha,numsets) abs ([],lf,ps,ns,f,rf,rs) =
  let fa ([],t,set) = t==abs
      ps'   = filter fa ps
      ns'   = filter fa ns
      rs'   = filter fa rs
      labs1 = ["ABS"++abs++(show i)| i<-[1..numsets]]
      labs2 = foldr (\([],t,set) la  -> seti la (alpha set) "1") labs1 ps'
      labs  = foldr (\([],t,set) la  -> seti la (alpha set) "0") labs2 ns'
      rabs1 = if abs `elem` f then ["0"| i <-[1..numsets]] 
      	      else ["ABS"++abs++(show i)| i<-[1..numsets]]
      rabs2 = foldr (\([],t,set) ra -> seti ra (alpha set) "0") rabs1 (ps'++ns')
      rabs  = foldr (\([],t,set) ra -> seti ra (alpha set) "1") rabs2 rs'
      lval  = (PComp "val" (map PAtom labs))
      rval  = (PComp "val" (map PAtom rabs))
      lf'   = subfs (Map.singleton abs lval) lf
      rf'   = subfs (Map.singleton abs rval) rf
      tir   = if lval==rval || abs `elem` f then [] else [("timplies",[lval,rval])]
  in ([],lf',ps,ns,f,tir++rf',rs)

seti :: [a] -> Int -> a -> [a]
seti list i v = let (l1,l2)=splitAt (i-1) list
                in l1++[v]++(tail l2)

absrs tab = concatMap $ absr tab
absr :: Symbtab -> PRule -> [PIdent]
absr  tab ([],lf,ps,ns,f,rf,rs) = 
  nub $
  (absfs tab lf)++
  (abscs tab ps)++
  (abscs tab ns)++
  f++
  (absfs tab rf)++
  (abscs tab rs)
absfs tab = concatMap $ absf tab
absf  tab (f,ts) = absts tab ts
abscs tab = concatMap $ absc tab
absc  tab ([],t,set) = [t]
absts tab = concatMap $ abst tab
abst :: Symbtab -> PTerm -> [PIdent]
abst  tab (PAtom a) = if isVal (tab ! a) then [a] else []
abst  tab (PComp f ts) = absts tab ts
abst  tab (PParId a) = error ("error in PParId " ++ a)
abst  tab (PComp f ts) = error ("error in PPar " ++ f)

getSetsrs :: [PRule] -> [PTerm]
getSetsrs = concatMap getSetsr
getSetsr (_,_,ps,ns,_,_,rs) =
  (getSetscs ps)++(getSetscs ns)++(getSetscs rs)
getSetscs = map (\ ([],t,set) -> set)

------- pretty printer for SPASS

pASLan :: Symbtab -> Int -> String -> [PRule]  -> String
pASLan tab numsets name rules = --- (name,types,sets,pubfun,privfun,facts,rules) =
  let sym = (Map.toList tab) in
  "begin_problem("++name++").\n"++
  "list_of_descriptions.\n"++
  "name({*"++name++"*}).\n"++
  "author({*FPASLan-to-SPASS compiler*}).\n"++
  "status(unknown).\n"++
  "description({*N/A*}).\n"++
  "end_of_list.\n\n"++
  "list_of_symbols.\n"++
  "functions["++
  (pX "," pSym (("val",FunSym False numsets): [(name,f) | (name,f)<-sym, isFun f || isEnuCo f]))++
  "].\n"++
  "predicates["++
  (pX "," pSym ([(name,f) | (name,f)<-sym, isEnu f || isFact f]))++
  "].\n"++
  "end_of_list.\n\n"++
  "list_of_formulae(axioms).\n"++
  (pX "\n" pRule rules)++"\n"++
  "%occs \n"++
  (pX "\n" pRule (val_occurrences tab rules))++"\n"++
  "end_of_list.\n\n"++
  "list_of_formulae(conjectures).\n"++
  "formula(attack).\n"++
  "end_of_list.\n"++
  "end_problem.\n"

val_occurrences :: Symbtab -> [PRule] -> [PRule]
val_occurrences tab rules =
  nub
     [ ([],("timplies",[PAtom "X", PAtom "Y"]):
       	   [(subposf f p (PAtom "X"))],
	[],[],[],
        [(subposf f p (PAtom "Y"))],[]) 
     | (f,poses) <- nub $ getOccrs tab rules, p<-poses ] 

type Position = [Int]

getOccrs :: Symbtab -> [PRule] -> [(PFact,[Position])]
getOccrs tab = concatMap (getOccr tab)

getOccr :: Symbtab ->  PRule -> [(PFact,[Position])]
getOccr tab ([],lf,_,_,_,rf,_) =
  [ (f,posVal tab f) | f<-lf++rf]

posVal :: Symbtab -> PFact -> [Position]
posVal tab ("timplies",_) = []
posVal tab (f,ts) = 
  nub [ (i:p) | (i,t)<- zip [1..] ts, p<- posValt tab [] t]

posValt :: Symbtab -> Position -> PTerm -> [Position]
posValt tab p (PAtom _) = []
posValt tab p (PComp "val" ts) = [p]
posValt tab p (PComp f ts) = concatMap (\ (i,t) -> posValt tab (p++[i]) t ) (zip [1..] ts)

subposf :: PFact -> Position -> PTerm -> PFact
subposf (f,ts) (i:p) t = 
  (f,seti ts i (subpost (head $ drop (i-1) ts) p t))

subpost :: PTerm -> Position -> PTerm -> PTerm
subpost _ [] t' = t'
subpost (PComp f ts) (i:p) t' = PComp f (seti ts i (subpost (head $ drop (i-1) ts) p t'))
subpost t p t' = error $ (show t)++"\n"++(show p)++"\n"++(show t')

pRule :: PRule -> String
pRule ([],lf,_,_,_,rf,_) = 
  let vars = nub $ fvfs(lf)++fvfs(rf)
      formula = if null lf then pFacts rf else
      	      	"implies("++(pFacts lf)++",\n"++
       	        (pFacts rf)++")"
      quantifier = if null vars then formula 
      		   else "forall(["++(pX "," (\x-> x) vars)++"],"++formula++")"
  in "formula("++quantifier++").\n"

pFacts [] = error "Empty RHS"
pFacts [x] = pFact x
pFacts (x:xs) = "and("++(pFact x)++","++(pFacts xs)++")"

pSym :: (PIdent,SymbEntry) -> String
pSym (f,FunSym _ n) = "("++f++","++(show n)++")"
pSym (c,EnuCo) = "("++c++",0)"
pSym (x,VarEnu _) = "("++x++",1)"
pSym (f,FactSym n) = "("++f++","++(show n)++")"

pX :: String -> (a->String) -> [a] -> String
pX sep f [] = ""
pX sep f [x] = f x
pX sep f xs = foldl (\ x y -> x++sep++(f y)) (f (head xs)) (tail xs)

pTerms = pX "," pTerm

pTerm (PAtom id) = id
pTerm (PComp id ts) = id++"("++(pTerms ts)++")"

pFact (id,ts) =  if (length ts)==0 then id else id++"("++(pTerms ts)++")"

{-
  "Types: "++(pX pTypedec types)++"\n"++
  "Sets: "++(pTerms sets)++"\n"++
  "Symbols: "++(pX pSymbol (pubfun++privfun++facts))++"\n"++
  "Rules: "++(pX pRule rules)++"\n"

pL = pX (\x -> x)

pTypedec (id,typ) = (pL id)++":"++(pTyp typ)++";"

pTyp (Enum list) = "{" ++ (pL list) ++ "}"
pTyp (Value) = "value"
pTyp (Untyped) = "untyped"

pSymbol (id,i) = id++"/"++(show i)

pRule :: PRule -> String
pRule (lambda,lf,ps,ns,f,rf,rs) =
  (if null lambda then "" else "\\"++(pL lambda)++".\n")++
  (if null lf then "" else (pX pFact lf)++"\n")++
  (if null ps then "" else (pX pCond ps)++"\n")++
  (if null ns then "" else (pX pNCond ns)++"\n")++
  "=["++(pL f)++"]=>\n"++
  (if null rf then "" else (pX pFact rf)++"\n")++
  (if null rs then "" else (pX pCond rs)++"\n")++"\n"
pCond ([],el,set) = el++" in "++(pTerm set)
pCond (al,el,set) = error "Allquantified positive condition" 
pNCond ([],el,set) = el++" notin "++(pTerm set)
pNCond (al,el,set) = "forall "++(pL al)++"."++el++" notin "++(pTerm set)

-}

------- pretty printer for ProVerif

prASLan :: Symbtab -> Int -> String -> [PRule]  -> String
prASLan tab numsets name rules = 
  let sym = (Map.toList tab) in
  "pred iknows/1 elimVar,decompData.\n"++
  (concatMap (\x -> "pred "++(prSym x)++".\n")
   ([(name,f) | (name,f)<-sym, isEnu f || isFact f, name/="iknows"]))++
  "nounif iknows:x.\n"++
  "fun nothing/0.\n"++
  (concatMap (\x -> "fun "++(prSym x)++".\n")
   (("val",FunSym False numsets):
    [(name,f) | (name,f)<-sym, isFun f, (FunSym _ n) <- [f], n>0]))++
  "query attack:.\n"++
  "reduc\n\n"++
  (pX "\n" prRule rules)++"\n"++
  "(* occs *) \n"++
  (pX "\n" prRule (val_occurrences tab rules))++"\n"++
  "iknows:nothing."

prRule :: PRule -> String
prRule ([],lf,_,_,_,rf,_) = 
  let vars = nub $ fvfs(lf)++fvfs(rf)
  in concat [ (if null lf then prFact r 
      	       else (prFacts lf)++" -> "++(prFact r))++";\n"
            | r <- rf ]

prFacts [] = error "Empty RHS"
prFacts [x] = prFact x
prFacts (x:xs) = (prFact x)++" & "++(prFacts xs)

prSym :: (PIdent,SymbEntry) -> String
prSym (f,FunSym _ n) = f++"/"++(show n)
prSym (c,EnuCo) = error "Co no in pv"
prSym (x,VarEnu _) = x++"/1"
prSym (f,FactSym n) = f++"/"++(show n)


prTerms = pX "," prTerm

prTerm (PAtom "0") = "Num0[]"
prTerm (PAtom "1") = "Num1[]"
prTerm (PAtom id) = if (head id) `elem` ['a'..'z'] then "C"++id++"[]" else "v"++id
prTerm (PComp id ts) = id++"("++(prTerms ts)++")"

prFact (id,ts) =  id++":"++(prTerms ts)
        
-- <paolo>
-- definitions translation and "occurs"

mkPrepros :: PASLan -> PASLan
mkPrepros (name,types,sets,pubfun,privfun,facts,definitions,rules) = let 
                                                                                                         definitions1 = mkDefinitions definitions definitions                           
                                                                                                    in (name,types,sets,pubfun,privfun,mkFacts facts,definitions1, mkOccurs . deParRules . mkDefinitions rules $ definitions1) 

-- "occurs"

occursfun :: String
occursfun = "occurs"

mkFacts :: [(PIdent,Int)] -> [(PIdent,Int)]
mkFacts x = x ++ [(occursfun,1)]

occurs :: PIdent -> PFact
occurs idf = (occursfun,[(PAtom idf)])

mkOccurs :: [PRule]->[PRule]
mkOccurs xs = map mkOccur xs

mkOccur :: PRule -> PRule
mkOccur (lamvar,lf,ps,ns,f,rf,rs) = let 
                                                        -- ident in term in the left side
                                                        lf1 = lf ++ (map (\x -> occurs x)) (nub ((getCondIds ps)) \\ f)
                                                        -- fresh terms
                                                        rf1 = rf ++ (map (\x -> occurs x)) (nub f)
                                                    in (lamvar,lf1,ps,ns,f,rf1,rs) 

getCondIds :: [PCond] -> [PIdent]
getCondIds [] =[]
getCondIds (([],id,_):[]) = [id]
getCondIds (x:xs) = getCondIds [x] ++ getCondIds xs

-- definitions translation

mkDefinitions :: Def a => [a] -> PDefinitions -> [a]
mkDefinitions a [] = a
mkDefinitions [] _ = []
mkDefinitions a defs = map (\x -> mkDefs(x) defs) a

mkDefs :: Def a => a -> PDefinitions -> a     
mkDefs a [] = a
mkDefs a (def:[]) = mkDef [def] a
mkDefs a (def:defs) = mkDef [def] (mkDefs a defs)

class Def a where
        mkDef :: PDefinitions -> a -> a

instance Def PRule where
        mkDef def (lamvar,lf,ps,ns,f,rf,rs) = (lamvar,mkDef def lf,ps,ns,f,mkDef def rf,rs)

instance Def [PRule] where
        mkDef def xs = map (mkDef def) xs
        
instance Def [PTerm] where
        mkDef def xs = map (mkDef def) xs

instance Def [PDefinition] where
        mkDef _ [] = []
        mkDef def (x:[]) = [mkDef def x]
        mkDef def (x:xs) = mkDef def (mkDef def x):xs 

instance Def PDefinition where
        mkDef def1 (idf,def,pars) = (idf,mkDef def1 def,pars)

instance Def [PFact] where
        mkDef def xs = map (mkDef def) xs

instance Def [PCond] where
        mkDef def xs = map (mkDef def) xs
                
instance Def PFact where
        mkDef def (ident,terms) = (ident,mkDef def terms) 
        
instance Def PCond where
        mkDef def (idents,ident,term) = (idents,ident, mkDef def term)

instance Def PTerm where
        mkDef ((idf,def,_):[]) (PAtom ident) = if ident==idf then def else PAtom ident
        mkDef _ (PComp ident []) = PComp ident []
        mkDef (def@(idf,defs,xds):[]) (PComp ident xs) = if ident==idf && length(xds)==length(xs) then trDefPar defs xds xs else PComp ident (mkDef [def] xs)
        mkDef ((idf,defs,xds):[])  (PPar ident xps) = if ident==idf && length(xds)==length(xps) then trDefPar defs xds xps else PPar ident xps
        mkDef _ (PParId xps) = PParId xps
        mkDef def term = error(show def ++ " definition not applicable on " ++ show term)

deParRule :: PRule -> PRule
deParRule (lamvar,lf,ps,ns,f,rf,rs) = (lamvar,deParFacts lf,ps,ns,f,deParFacts rf,rs)

deParRules :: [PRule] -> [PRule]
deParRules xs = map deParRule xs  

deParFacts :: [PFact] -> [PFact]
deParFacts xs = map deParFact xs 

deParFact :: PFact -> PFact
deParFact f@(_,[]) = f
deParFact (ident,terms) = (ident,dePars terms)

dePar :: PTerm -> PTerm
dePar t@(PAtom _) = t
dePar t@(PComp _ []) = t
dePar (PParId idf) = PAtom idf
dePar (PComp ident (x:xs)) = PComp ident (dePar x : dePars xs)

dePar (PPar ident pars) = PComp ident pars

dePars :: [PTerm] -> [PTerm] 
dePars xs = map dePar xs

trDefPars :: [PTerm] -> [PIdent] -> [PTerm] -> [PTerm]
trDefPars xs xds xps = map (\x -> trDefPar x xds xps) xs

trDefPar :: PTerm -> [PIdent] -> [PTerm] -> PTerm
trDefPar (PPar ident ids) xds xps = PPar ident [snd x | x <-zip xds xps, elem (PAtom (fst x)) ids]
trDefPar (PComp ident (x:xs)) xds xps = PComp ident ([trDefPar x xds xps]++trDefPars xs xds xps)
trDefPar (PParId ident) xds xps = if (elem ident xds) then (head [snd x | x <-zip xds xps, elem (fst x) [ident]]) else PParId ident
trDefPar (PAtom ident) _ _  = PParId ident
trDefPar t _ _ = t

--------------- debug -----------------------------
dbgDefs  :: PASLan -> String
dbgDefs (paslan@(name,types,sets,pubfun,privfun,facts,definitions,rules)) = 
        let 
             str = "\n\n--- Definitions ---\n\n" ++ printDefs definitions ++ "\n--- Rules ---\n\n" ++ printRules rules
             p1@(name,types,sets,pubfun,privfun,facts,definitions1,r1) = mkPrepros paslan
             str1 = "\n--- Definitions ---\n\n" ++ printDefs definitions1 ++ "\n--- Rules Changed ---\n\n" ++ printRules r1 ++ "\n--- Protocol Translation ---\n\n"
         in  (str ++ str1)
              
printDefs :: PDefinitions -> String
printDefs ([]) = show ""
printDefs (x:[]) = printDef x
printDefs (x:xs) = printDef x ++ printDefs xs

printDef :: PDefinition -> String
printDef (idf,def,pars) = show idf ++ "(" ++ show pars ++ "): " ++ show def ++ "\n"

printRules :: [PRule] -> String
printRules ([]) = show ""
printRules (x:[]) = printRule x
printRules (x:xs) = printRule x ++ printRules xs
        
printRule :: PRule -> String
printRule rule = show rule ++ "\n"

--------------------------------------------------

-- </paolo>
