{

{-

Open Source Fixedpoint Model-Checker version 2010

(C) Copyright IBM Corp. 2010

All Rights Reserved.

Modifications by Paolo Modesti 2010

-}

module FPASLanParser where
import Char
import List
import FPASLanLexer
}


%name parser
%tokentype {Token}

%token	    ident		{Tident _ $$}
	    int			{Tint _ $$}

	    'Problem:' {Tproblem _}
	    'Types:' {Ttypes _}
	    'Sets:' {Tsets _}
	    'Functions:' {Tfunctions _}
	    'public' {Tpublic _}
	    'private' {Tprivate _}
	    'Facts:' {TFacts _}
	    'Rules:' {TRules _}
	   	'Definitions:' {TDefinitions _}
	    'value' {Tvalue _}
	    'untyped' {Tuntyped _}
	    'notin' {Tnotin _}
	    'in' {Tin _}
	    'forall' {Tforall _}
	    'exists' {Texists _}
	    ';' {Tsemicolon _}
	    ':' {Tcolon _}
	    ',' {Tcomma _}
	    '{' {Topenbrace _}
	    '}' {Tclosebrace _}
	    '[' {Topensquare _}
	    ']' {Tclosesquare _}
	    '(' {Topenpar _}
	    ')' {Tclosepar _}
	    '/' {Tslash _}
	    '\\' {Tlambda _}
	    '.' {Tdot _}
	    '=>' {Tarrow _}
	    '=[' {Topenarrow _}
	    ']=>' {Tclosarrow _}


%%


FPASLan :: {PASLan}
: 'Problem:'   ident ';'
  'Types:'     typedecs
  'Sets:'      terms ';'
  'Functions:' 'public' symdecs ';'
               'private' symdecs ';'
  'Facts:'     symdecs ';'
-- <paolo>
    defdec
-- </paolo>
  'Rules:'     rules
{($2,$5,$7,$11,$14,$17,$19,$21)}

-- <paolo>
defdec :: {PDefinitions}
: {[]}
|'Definitions:' deflist {$2}

deflist :: {PDefinitions}
: ident ':' term {[($1,$3,[])]}
| ident '[' idents ']' ':' term {[($1,$6,$3)]}
| ident ':' term ';' deflist {(($1,$3,[]):$5)}
| ident '[' idents ']' ':' term ';' deflist {(($1,$6,$3):$8)}

-- </paolo>

typedecs :: {[Typedec]}
: idents ':' type ';' typedecs {(($1,$3):$5)}
| {[]}

idents :: {[PIdent]}
: ident {[$1]}
| ident ',' idents {($1:$3)}

type :: {PType}
: '{' idents '}' {Enum $2}
| 'value' {Value}
| 'untyped' {Untyped}

terms :: {[PTerm]}
: term {[$1]}
| term ',' terms {($1:$3)}

term :: {PTerm}
: ident  {PAtom $1}
| ident '(' terms ')' {PComp $1 $3}
| ident '[' terms ']' {PPar $1 $3}
--<paolo>
| '[' ident ']' {PParId $2}
--</paolo>
symdecs :: {[(PIdent,Int)]}
: ident '/' int {[($1,$3)]}
| ident '/' int ',' symdecs {($1,$3):$5}

rules :: {[PRule]}
: rule rules {$1:$2}
| {[]}

rule :: {PRule}
: lambda lhss arrow rhss ';'
{  (\ (lf,ps,ns) (rf,rs,_) -> ($1,lf,ps,ns,$3,rf,rs)) $2 $4 
}

lambda :: {[PIdent]}
: '\\' idents '.' {$2}
| {[]}

lhss :: {([PFact],[PCond],[PCond])}
: lhs '.' lhss {combside $1 $3}
| {([],[],[])}
| lhs {$1}

rhss :: {([PFact],[PCond],[PCond])}
: rhs '.' rhss {combside $1 $3}
| {([],[],[])}
| rhs {$1}

lhs:: {([PFact],[PCond],[PCond])}
: rhs {$1}
| ident 'notin' term {[],[],[([],$1,$3)]}
| 'forall' idents '.' ident 'notin' term {[],[],[($2,$4,$6)]}

rhs:: {([PFact],[PCond],[PCond])}
: pfact {([$1],[],[])}
| ident 'in' term {[],[([],$1,$3)],[]}

pfact:: {PFact}
: ident '(' terms ')' {($1,$3)}
| ident  {($1,[])}

arrow :: {[PIdent]}
: '=>' {[]}
| '=[' idents ']=>' {$2}


{
happyError :: [Token] -> a
happyError tks = error ("Parse error at " ++ lcn ++ "\n" )
	where
	lcn = 	case tks of
		  [] -> "end of file"
		  tk:_ -> "line " ++ show l ++ ", column " ++ show c ++ " - Token: " ++ show tk
			where
			AlexPn _ l c = token_posn tk

type PASLan = (PIdent,[Typedec],[PTerm],
	       [(PIdent,Int)],[(PIdent,Int)],[(PIdent,Int)],PDefinitions,[PRule])
type Typedec = ([PIdent],PType)
data PType = Enum [PIdent]
           | Value
           | Untyped
           deriving (Eq,Show)
data PTerm = PAtom PIdent
           | PComp PIdent [PTerm]
-- <paolo>
           | PPar PIdent [PTerm]
           | PParId PIdent
-- </paolo>
           deriving (Eq,Show,Ord)
type PFact = (PIdent,[PTerm])
type PCond = ([PIdent],PIdent,PTerm)
type PRule=([PIdent],[PFact],[PCond],[PCond],[PIdent],[PFact],[PCond])
-- <paolo>
type PDefinition = (PIdent,PTerm,[PIdent])
type PDefinitions = [PDefinition]
-- </paolo>
combside :: ([PFact],[PCond],[PCond]) -> ([PFact],[PCond],[PCond]) -> ([PFact],[PCond],[PCond]) 
combside (pf,ps,ns) (pf',ps',ns') = (pf++pf',ps++ps',ns++ns')
 }
